/* 
 * Author: Yueping Zhang, Internet Research Lab (IRL), Texas A&M University
 * Email: yueping@cs.tamu.edu
 * File Name: jetmax-queue.cc
 * Functionality: implementation of the JetMax router
 * Last modified on Dec. 5, 2005
 */

#include "jetmax-queue.h"

/*
 * Delare the class of the JetMax router
 */
static class JetMaxQueueClass : public TclClass {
public:
	JetMaxQueueClass() : TclClass("Queue/JetMaxQueue") {}
	TclObject* create(int, const char*const*) {
		return (new JetMaxQueue);
	}
} class_JetMax_queue;

/*
 * Constructor of the JetMax router class
 */
JetMaxQueue::JetMaxQueue()
{
  	double now=Scheduler::instance().clock();
  	q_ = new PacketQueue();	
  	
	// bind Tcl variables
	bind("C",&JetMaxp_.C);
	bind("T",&delta_);
	bind("qID", &qID_);
	bind("gamma_", &gamma_);
	
	// initialize local variables
  	JetMaxp_.rate_total = 0;
	JetMaxp_.pkt_loss = 2;
  	JetMaxp_.congested = 0;
  	JetMaxp_.unresponsive = 0;
	JetMaxp_.N_congested = 0;
	JetMaxp_.rid_c = -1;
	JetMaxp_.g = -1;

  	t_finish_ = now + delta_;
  	t_start_ = now;
	duration_ = delta_;
	approval_ = 0;
	cur_interval_ = 1;	
  	is_first_pkt_ = 1;
}

/*
 * Receive function, called upon each packet arrival
 */
void JetMaxQueue::recv(Packet* p, Handler* h){
  	hdr_cmn* ch = hdr_cmn::access(p);
	double now = Scheduler::instance().clock();

	// if the incoming packet is a JetMax data packet
 	if(ch->ptype()==PT_JetMax)
	{ 
  		hdr_JetMax* header_jetmax;	
		header_jetmax = hdr_JetMax::access(p);
		if (header_jetmax->ack == 0)
		{
			if(p!=0)
    			Queue::recv(p,h);	

			compute_ack(header_jetmax);
			do_on_packet_arrival(header_jetmax);
			write_packet(header_jetmax);
		}
		else
		{ 
			if(p!=0)
    			Queue::recv(p,h);	
		}
	}
}

/*
 * Operations conducted upon the arrival of each packet 
 */
void JetMaxQueue::do_on_packet_arrival(hdr_JetMax* header_jetmax)
{
	double now = Scheduler::instance().clock();
	if(is_first_pkt_){
		// record the starting time of the current sampling interval
  		t_start_ = now;
  		t_finish_ = now + delta_;
  		is_first_pkt_ = 0;
		cur_interval_ = header_jetmax->interval;
		JetMaxp_.C *= gamma_;
  	}	

	// record and increment rid_c
	header_jetmax->rid_c++;
	JetMaxp_.rid_c = header_jetmax->rid_c;

	// check whether myself is the chosen bottleneck router
	if(header_jetmax->rid_t == JetMaxp_.rid_c)
	{
		JetMaxp_.congested += header_jetmax->psize_proposed;
		// estimate the number of congested flows
		JetMaxp_.N_congested += header_jetmax->delta;
		header_jetmax->interval = cur_interval_;
	}
	// if myself is not the chosen bottleneck router, count the incoming flow as unresponsive flow
	else
		JetMaxp_.unresponsive += header_jetmax->psize_proposed;
}


/* 
 * Compute the feedback 
 */
void JetMaxQueue::compute_ack(hdr_JetMax* header_jetmax)
{
	double now=Scheduler::instance().clock();
	if(now >= t_finish_ && now > t_start_)
	{	
		duration_ = now - t_start_;

		cur_interval_ ++;

		// calculate the incoming rate during the last interval delta
		JetMaxp_.rate_total = (JetMaxp_.congested + JetMaxp_.unresponsive) / duration_;

		// calculate the packet loss
		if (JetMaxp_.rate_total != 0) 
			JetMaxp_.pkt_loss = 1 - JetMaxp_.C / JetMaxp_.rate_total; 
		else
			JetMaxp_.pkt_loss = 2; 
		
		// calculate the avergare congested rate
		JetMaxp_.N_congested = JetMaxp_.N_congested / duration_;

		if (JetMaxp_.N_congested > 0)
			JetMaxp_.g = (JetMaxp_.C - JetMaxp_.unresponsive / duration_) / JetMaxp_.N_congested;
		else
			JetMaxp_.g = 0;

		// decide whether approve the proposed rate
		if (JetMaxp_.rate_total > JetMaxp_.C)
			approval_ = 0;
		else
			approval_ = 1;

		// reset the variables for the next control interval
		t_start_ = now;
		t_finish_ = t_start_ + delta_; 
   		JetMaxp_.unresponsive = 0;	
		JetMaxp_.congested = 0;
		JetMaxp_.N_congested = 0;
	}
	else if (now <= t_start_)
	{
		t_start_ = now;
		t_finish_ = t_start_ + delta_;
	}
}

/* 
 * Write feedback in the packet header 
 */
void JetMaxQueue::write_packet(hdr_JetMax* header_jetmax)
{	
	if (header_jetmax->approved != 0 && approval_ == 0)
		header_jetmax->approved = approval_;

	// if the packet loss of previous routers is smaller than that of myself,
	// or I am the first router in the path, overwrite fields "flost", "rid_s" and "g"
	double diff = fabs(header_jetmax->p - JetMaxp_.pkt_loss);
	if((header_jetmax->p < JetMaxp_.pkt_loss && diff > MIN_DISTANCE || header_jetmax->p == 2) && JetMaxp_.pkt_loss != 2
		&& header_jetmax->rid_s != 0  // rid_c == 0 means some router in the path is not ready
		)
	{
		if (JetMaxp_.g != 0)
			header_jetmax->g = JetMaxp_.g;

		header_jetmax->p = JetMaxp_.pkt_loss;
		header_jetmax->rid_s = JetMaxp_.rid_c;
	}
	// if there is a tie in the packet loss (both of which are zeros), switch to the
	// link with the minimum average congested rate
	else if (JetMaxp_.pkt_loss != 2 && diff < MIN_DISTANCE && is_loss_zero(header_jetmax) && header_jetmax->rid_s != 0)
	{
		// switch based on g
		if (less_than_min(header_jetmax))		
		{
			header_jetmax->g = JetMaxp_.g;
			header_jetmax->rid_s = JetMaxp_.rid_c;
			header_jetmax->p = JetMaxp_.pkt_loss;
		}
		else if (JetMaxp_.g <= 0)
			header_jetmax->p = 0;
	}

	// notify the sender that at least one router in its path is not ready
	if (JetMaxp_.pkt_loss == 2)
		header_jetmax->rid_s = 0;
} 

/* 
 * Return the top packet in the queue for transmission 
 */
Packet* JetMaxQueue::deque() 
{
	Packet *p = q_->deque();
	return (p);
}

/*
 * Place the packet at the tail of the queue
 */
void JetMaxQueue::enque(Packet* pkt)
{
	q_ -> enque(pkt);
}

/* 
 * Examine whether or not the local average congested rate is 
 * smaller than the one carried in the packet header 
 */
bool JetMaxQueue::less_than_min(hdr_JetMax* header_jetmax)
{
	if (header_jetmax->g > JetMaxp_.g
		&& fabs(header_jetmax->g - JetMaxp_.g) > MIN_DISTANCE
		&& JetMaxp_.g != 0 || header_jetmax->g == -1)
		return(true);
	else
		return(false);
}

/* 
 * Examine whether or not the packet loss is zero 
 */
bool JetMaxQueue::is_loss_zero(hdr_JetMax* header_jetmax)
{
	if (fabs(header_jetmax->p) < MIN_DISTANCE && fabs(JetMaxp_.pkt_loss) < MIN_DISTANCE)
		return(true);
	else
		return(false);
}