/* 
 * Author: Yueping Zhang, Internet Research Lab (IRL), Texas A&M University
 * Email: yueping@cs.tamu.edu
 * File Name: jetmax-sink.cc
 * Functionality: implementation of the JetMax receiver
 * Last modified on Nov. 15, 2005
 */

#include "jetmax-sink.h"

/*
 * Delare the class of the JetMax receiver
 */
static class JetMaxSinkClass : public TclClass {
public:
  	JetMaxSinkClass() : TclClass("Agent/JetMaxSink") {}
  	TclObject* create(int, const char*const*) {
     		return (new JetMaxSinkAgent());
  	}
} class_JetMaxSink; 

/*
 * Constructor of an JetMax receiver
 */
JetMaxSinkAgent::JetMaxSinkAgent() : Agent(PT_JetMax), ack_timer(this)
{
	bind("packetSize_", &size_); // packet size of the ACK

	head = 0;
	tail = 0;
	next_ack_time = 0.001; // default value is 10 ms
	interval_num = 50;
	cur_num = 0;
#ifdef RANDOM_DELAY
	ackQueue = (ackQ *) malloc(QSIZE * sizeof(ackQ));
	for (int i = 0; i < QSIZE; i++)
		ackQueue[i].isEmpty = 1;
#endif
	isFirstPkt = 1;
}

/*
 * Receive function, called upon each packet arrival
 */
void JetMaxSinkAgent::recv(Packet *pkt, Handler *)
{
	hdr_JetMax *JetMaxh = hdr_JetMax::access(pkt); 

#ifdef RANDOM_DELAY
	// insert random delay into the reverse path 
	if (isFirstPkt == 1)
	{
		// send the first ACK and schedule the time for the next ACK
		sendpkt(pkt);
		Packet::free(pkt);
		isFirstPkt = 0;
	}
	else 
		AckEnque(pkt); // place the packet at the tail of the queue
#else
	// send the ACK immediately
	sendpkt(pkt);
	Packet::free(pkt);
#endif
}


/*
 * Initialize and send out the ACK
 */
void JetMaxSinkAgent::sendpkt(Packet *in_pkt)
{
	Packet* ack_pkt = allocpkt();
	if (ack_pkt == NULL) {
		printf  ("error allocating packet\n");
		abort(); 
	}

	hdr_JetMax *JetMax_ackh = hdr_JetMax::access(ack_pkt);
	hdr_JetMax *JetMaxh = hdr_JetMax::access(in_pkt);

	// copy feedbacks from the incoming packet to the ACK
	JetMax_ackh->p		= JetMaxh->p;
	JetMax_ackh->pkt_seqno	= JetMaxh->pkt_seqno;
	JetMax_ackh->rid_t	= JetMaxh->rid_t;
	JetMax_ackh->rid_c	= JetMaxh->rid_c;
	JetMax_ackh->rid_s = JetMaxh->rid_s;
	JetMax_ackh->psize_proposed = JetMaxh->psize_proposed;
	JetMax_ackh->approved = JetMaxh->approved;
	JetMax_ackh->ack = 1;
	JetMax_ackh->interval = JetMaxh->interval;
	JetMax_ackh->g = JetMaxh->g;

	send(ack_pkt,0);
	Packet::free(in_pkt);

#ifdef RANDOM_DELAY
	// schedule the next ACK
	ack_timer.resched(next_ack_time);
#endif
}

/*
 * Define TCL commands
 */
int JetMaxSinkAgent::command(int argc, const char*const* argv) 
{
	return (Agent::command(argc,argv));
}

/*
 * Enque function of the local ACK queue
 */
void JetMaxSinkAgent::AckEnque(Packet *pkt)
{	
	ackQueue[tail].isEmpty = 0;
	ackQueue[tail].pkt = pkt;

	tail = (tail + 1) % QSIZE;
	if (tail == head)
	{
		printf("Insufficient queue size!\n");
		exit(0);
	}
}

/*
 * Deque function of the local ACK queue
 */
Packet* JetMaxSinkAgent::AckDeque()
{
	if (ackQueue[head].isEmpty == 0)
	{
		ackQueue[head].isEmpty		= 1;
		int temp_head = head;
		head = (head + 1) % QSIZE;
		return(ackQueue[temp_head].pkt);
	}
	else
		return(NULL);
}

/*
 * Every time the timer expires, generate a delay spike
 */
void JetMaxAckTimer::expire(Event *) {
	if (a_->cur_num >= a_->interval_num)
	{
		a_->interval_num = (rand() % 5000 + 5000);
		a_->cur_num = 0;
		a_->next_ack_time = (rand() % 500 + 500) / 1000.0;
		int qlen;
		if (a_->head > a_->tail)
			qlen = a_->tail + QSIZE - a_->head;
		else
			qlen = a_->tail - a_->head;
	}
	else
	{
		a_->next_ack_time = 1e-5;
		a_->cur_num ++;
	}

	Packet *pkt = a_->AckDeque();
	if (pkt != NULL)
		a_->sendpkt(pkt);
	else
		a_->ack_timer.resched(a_->next_ack_time);
}

