/* 
 * Author: Yueping Zhang, Internet Research Lab (IRL), Texas A&M University
 * Email: yueping@cs.tamu.edu
 * File Name: jetmax-source.cc
 * Functionality: implementation of the JetMax sender
 * Last modified on Nov. 15, 2005
 */

#include "jetmax-source.h"

int hdr_JetMax::offset_;

/*
 * JetMax packet header class
 */
static class JetMaxHeaderClass : public PacketHeaderClass {
public:
	JetMaxHeaderClass() : PacketHeaderClass("PacketHeader/JetMax",
	sizeof(hdr_JetMax)) {
		bind_offset(&hdr_JetMax::offset_);
	}
} class_JetMaxhdr;

/*
 * JetMax sender agent class
 */
static class JetMaxClass : public TclClass {
public:
  	JetMaxClass() : TclClass("Agent/JetMaxSource") {}
  	TclObject* create(int, const char*const*) {
		return (new JetMaxSourceAgent());
	}
} class_JetMax;

/*
 * Constructor of the JetMax sender
 */
JetMaxSourceAgent::JetMaxSourceAgent() : Agent(PT_JetMax), send_timer_(this)
{	
	// bind TCL variables
  	bind("packetSize_", &psize_);  // fixed packet size, defined in Agent.h
  	bind("totalRate_", &rate_);	   // current sending rate
  	bind("InitRate_", &init_rate_);// initial rate
	bind("tau_",&tau_);	           // control gain
	bind("fid", &fid_);            // flow ID
}


/* 
 * Start the JetMax sender
 */
void JetMaxSourceAgent::start()
{
	double now = Scheduler::instance().clock();	

	rate_ = init_rate_;	           // set sending rate to the initial rate
	rate_proposed_ = init_rate_;   // set proposed rate to the initial rate
	psize_ *= 8.0;                 // packet size in bits
	psize_proposed_ = psize_;      // set proposed packet size to the actual value
	cur_pkt_seqno_ = 0;            // packet sequence number starts from zero
	is_rtt_expired_ = 0;           // initially RTT timer is not set
	seqno_last_rtt = 0;            // set it to the sequence number of the first packet
  	max_loss_seqno_ = -1;          // no feedback received yet
	cur_loss_seqno_ = -1;          // as above
	max_pkt_seqno_ = -1;           // as above
	pkt_loss_ = 2;                 // as above
	current_ACK_pkt_seqno_ = 0;    // as above
	switch_in_progress_ = 0;       // no bottleneck switching is in progress
	test_in_progress_ = 0;         // as above
	test_switch_ = 0;              // as above
	interval_cnt_ = 3;             // delay response to bottleneck switching by 2 delta 
	is_confirmed_ = 1;             // don't need to delay response to the first ACK
	cur_interval_ = 1;             // default value
	rid_t = -1;                    // bottleneck router has not been decided
	stationary = 0;                // not in steady state
	active_ = 1;                   // start the JetMax sender

	printf("Flow %d starts at time %f\n", fid_, now);

	// schedule the first packet to be sent
	send_timer_.resched(psize_ / rate_); 
}

/*
 * Stop the JetMax sender
 */
void JetMaxSourceAgent::stop()
{
	active_ = 0;
	send_timer_.force_cancel();
	rate_ = 0;
}

/*
 * Send out one packet and schedule the next one
 */
void JetMaxSourceAgent::nextpkt()
{
  	double next;
	if (active_ == 1)
	{
		if (rate_ != 0)
		{			
			sendpkt();// send a packet right now

			next = psize_ / rate_;    // inter-packet interval
  			send_timer_.resched(next);// schedule the next packet
		}
		else
		{
			printf("Sending rate is zero!\n");
			exit(0);
		}
	}
}

/*
 * Define TCL commands
 */
int JetMaxSourceAgent::command(int argc, const char*const* argv)
{
	if (argc==2) {
		if (strcmp(argv[1],"start")==0) {
			start();
			return TCL_OK;
		}
		if (strcmp(argv[1],"stop")==0) {
			stop();
			return TCL_OK;
		}
	}
  	return (Agent::command(argc, argv));
}

/* 
 * Receive function, called upon each ACK arrival
 */
void JetMaxSourceAgent::recv(Packet *pkt, Handler *)
{
	if (active_ == 1)
	{
		double now = Scheduler::instance().clock();
		hdr_JetMax *nck = hdr_JetMax::access(pkt);

		// if is a new valid ACK, increment the loss seqno, update the 
		// interval identifier, and record the seqno of the new ACK
		if (nck->pkt_seqno > max_pkt_seqno_ && nck->interval != cur_interval_ && nck->p != 2)
		{
			cur_loss_seqno_ ++;
			cur_interval_ = nck->interval;
			max_pkt_seqno_ = nck->pkt_seqno;
		}

		// Manage bottleneck switching:
		// in the progress of switching, delay response to feedbacks by one RTT 
		// + one Delta interval. This is accomplished by variable switch_in_progress: 
		// 0 -- no switching in progress, 1 -- start switching, 
		// 2 -- one RTT elpased, 3 -- one Delta interval elapsed
		if (nck->rid_s != rid_t && nck->rid_s > 0 && switch_in_progress_ == 0 && test_in_progress_ == 0)
			verify_switch(nck, now);
		else if (nck->rid_s == rid_t)
			test_switch_ = 0; // if switching doesn't persist, cancel the switching test

		// check if one RTT has elapsed since the bottleneck switching was detected
		if (switch_in_progress_ ==1 && nck->pkt_seqno >= seqno_last_rtt)
		{
			is_rtt_expired_ = 1; // update the signal
			max_loss_seqno_ = cur_loss_seqno_; // update loss sequence number
			switch_in_progress_ = 2; // change the status
		}
		
		// check if the packet loss is valid
		if(nck->p != 2 && nck->rid_s > 0)
		{	
			// check if one RTT has elapsed
			if (nck->pkt_seqno >= seqno_last_rtt && is_rtt_expired_ == 0)
				is_rtt_expired_ = 1;
			
			// don't forget to update packet loss sequence number during the RTT
			if (cur_loss_seqno_ > max_loss_seqno_ && is_rtt_expired_ == 0)
			{
				max_loss_seqno_ = cur_loss_seqno_;
				max_pkt_seqno_ = nck->pkt_seqno;
			}

			if(cur_loss_seqno_ > max_loss_seqno_ && is_rtt_expired_ == 1)
			{	
				max_loss_seqno_ = cur_loss_seqno_; // update the maximum loss sequence number

				// delay response by one more delta
				if (switch_in_progress_ == 2)
					switch_in_progress_ = 3;

				if (test_in_progress_ != 0)
					test_in_progress_ ++;

				if (switch_in_progress_ == 0 && test_in_progress_ == 0)
				{					
					test_switch_ = 0;		// test passed, reset the flag

					// retrieve information from the ACK packet header
					pkt_loss_ = nck->p;
					rate_avg_ = nck->g;
					max_pkt_seqno_ = nck->pkt_seqno;
					current_ACK_pkt_seqno_=nck->pkt_seqno;
					
					ref_rate_ = rate_proposed_;		// use the last proposed rate as the reference rate
					update_rate(nck,now);			// update the sending rate

					// check if reaches steady state
					double diff = fabs(rate_ - nck->g);
					if (diff < nck->g * 0.001)
						stationary = 1;
					else
						stationary = 0;
				}

				if (test_in_progress_ == interval_cnt_)
					test_in_progress_ = 0;

				if (switch_in_progress_ == 3)
					switch_in_progress_ = 0;

				// schedule the next packet
				nextpkt();	
			}
		}
		Packet::free(pkt);
	}
}


/* 
 * Adjust the sending rate based on the feedback 
 */
void JetMaxSourceAgent::update_rate(hdr_JetMax *nck, double now)
{
	is_confirmed_++;
	// wait for one more delta such that the previous proposed 
	// rate can be directly used as the reference rate
	if (is_confirmed_ >= 2)
	{
		// if the proposed rate is approved, apply it
		if (nck->approved != 0)
		{
			rate_ = rate_proposed_;

			rate_proposed_ = ref_rate_ - tau_ * (ref_rate_ - rate_avg_);

			// decreasing rate does not need approval
			if (rate_proposed_ <= rate_)	
				rate_ = rate_proposed_;

			// update the proposed packet size
			psize_proposed_ = psize_ * rate_proposed_ / rate_;

			is_confirmed_ = 0;
			seqno_last_rtt = cur_pkt_seqno_;
			is_rtt_expired_ = 0;
		}		
		// if the proposed rate is not approved, propose 
		// a new rate based on the current feedback
		else
		{			
			double tmp_rate_proposed_ = ref_rate_ - tau_ * (ref_rate_ - rate_avg_);

			// if the new proposed rate is smaller than the current rate, 
			// immediately apply it; otherwize, preserve the old preposed rate
			if (tmp_rate_proposed_ <= rate_)
			{
				rate_proposed_ = tmp_rate_proposed_;
				rate_ = rate_proposed_;	// directly apply this proposed rate

				// update the proposed packet size
				psize_proposed_ = psize_ * rate_proposed_ / rate_;

				is_confirmed_ = 0;
				seqno_last_rtt = cur_pkt_seqno_;
				is_rtt_expired_ = 0;
			}
			else if (tmp_rate_proposed_ < rate_proposed_)
			{
				rate_proposed_ = tmp_rate_proposed_;
				psize_proposed_ = psize_ * rate_proposed_ / rate_;
				is_confirmed_ = 0;
				seqno_last_rtt = cur_pkt_seqno_;
				is_rtt_expired_ = 0;
			}
		}		
	}
}

/*
 * Initialize and send out a packet
 */
void JetMaxSourceAgent::sendpkt()
{
  	if (active_) {

		Packet* p = allocpkt();
		hdr_JetMax *JetMaxh = hdr_JetMax::access(p);

		// write packet header
		JetMaxh->p = 2;
		JetMaxh->rid_t = rid_t;
		JetMaxh->rid_s = -1;	// rid_suggested = -1 means unchanged, rid_suggested = 0 means 
								// the feedback of some router in the path in not ready yet
		JetMaxh->rid_c = 0;     // router ID counter starts from zero

		// NOTE: delta is calculated based on the actual sending rate, not the proposed one
		if (rate_proposed_ > 0)
            JetMaxh->delta = psize_ / rate_;
		else
			JetMaxh->delta = 0;

		JetMaxh->psize_proposed = psize_proposed_;
		JetMaxh->g = -1;
		JetMaxh->approved = 1;
		JetMaxh->ack = 0;
		JetMaxh->interval = cur_interval_;
		JetMaxh->pkt_seqno = cur_pkt_seqno_;

		cur_pkt_seqno_++; // packet sequence number increments by one

		send(p, 0);
	}
}

/*
 * Schedule the next packet to be sent
 */
void JetMaxSendTimer::expire(Event *) 
{
  	a_->nextpkt();
}

/*
 * Manage bottleneck switching
 */
void JetMaxSourceAgent::verify_switch(hdr_JetMax *nck, double now)
{
	// switching test is waivered for new flow
	if (rid_t == -1)
	{
		rid_t = nck->rid_s;
		switch_in_progress_ = 1;
		seqno_last_rtt = cur_pkt_seqno_;
		is_rtt_expired_ = 0;
		printf("	Time %f, flow %d's bottleneck switches to %d\n", now, fid_, rid_t);		
	}
	else
	{
		// Switch based on packet loss when the difference between the old and new packet 
		// loss exceeds a certain threshold (0.01 in the current code). Otherwise, switch based
		// on average rate g as explained in the INFOCOM 06 paper
		double diff = fabs(nck->p - pkt_loss_);

		// switch based on p
		if (diff >= 0.01)
		{
			// when detect a bottleneck switch, delay the response by 
			// RTT + 2delta. If persists, apply this switch
			if (test_switch_ == 0)
			{				
				test_in_progress_ = 1;
				test_switch_ = 1;
				interval_cnt_ = 3;				
			}
			else
			{
				rid_t = nck->rid_s;
				test_switch_ = 0;
				switch_in_progress_ = 1;
				seqno_last_rtt = cur_pkt_seqno_;
				is_rtt_expired_ = 0;
				printf("	Time %f, flow %d's bottleneck switches to %d\n", now, fid_, rid_t);		
			}
		}
		// switch based on g
		else if (rate_ > nck->g && nck->g != -1 && nck->g != 0)
		{
			// to prevent bottleneck oscillation immediately after one switch, if switch based 
			// on g, wait until stationarity is reached under the current bottleneck assignment
			if (stationary != 0)
			{
				if (test_switch_ == 0)
				{
					test_in_progress_ = 1;
					test_switch_ = 1;
					interval_cnt_ = 3;				
				}
				else
				{
					rid_t = nck->rid_s;
					test_switch_ = 0;
					switch_in_progress_ = 1;
					seqno_last_rtt = cur_pkt_seqno_;
					is_rtt_expired_ = 0;
					printf("	Time %f, flow %d's bottleneck switches to %d\n", now, fid_, rid_t);
				}
			}
		}
	}
}