/* 
 * Author: Yueping Zhang, Internet Research Lab (IRL), Texas A&M University
 * Email: yueping@cs.tamu.edu
 * File Name: jetmax-source.h
 * Functionality: header file of the JetMax sender
 * Last modified on Nov. 15, 2005
 */

#include <math.h>
#include "agent.h"

/*
 * Define packet header format of JetMax
 */
struct hdr_JetMax {	
	int		pkt_seqno;	    // packet sequence number
	int		rid_c:8;        // ID of the current router
	int		rid_t:8;	    // ID of the bottleneck router
	int		rid_s:8;        // router ID suggested by a router in the path
	// flags
	char	ack:1;          // is the packet an ACK?
	char	approved:1;     // did the bottleneck router approved the proposed rate?
	char	interval:1;     // current loss intervel, used to infer loss sequence number
	float	psize_proposed;	// proposed packet size
	float	p;              // packet loss at the bottleneck router
	float	g;              // ideal average rate
	float	delta;          // control interval at the bottleneck router
	static int	offset_;    // offset for this header
	inline static int& offset() { 
		return offset_; 
	}

	inline static hdr_JetMax* access(const Packet* p) {
		return (hdr_JetMax*) p->access(offset_);
	}
};


class JetMaxSourceAgent;

/* 
 * Timer for scheduling out-going packets
 */
class JetMaxSendTimer : public TimerHandler {
public:
	JetMaxSendTimer(JetMaxSourceAgent *a) : TimerHandler() { a_ = a; }
	virtual void expire(Event *e);
protected:
	JetMaxSourceAgent *a_;
};	

/*
 * JetMax sender agent
 */
class JetMaxSourceAgent : public Agent 
{
	friend class JetMaxSendTimer;
public:
	JetMaxSourceAgent();
	int		command(int argc, const char*const* argv); 
	void	recv(Packet*, Handler*);// receive function, called upon each ACK arrival
	void	sendpkt();              // initialize the packet header and send out the packet
	void	nextpkt();              // schedule the next out-going packet
	void	update_rate(hdr_JetMax*, double); // adjust the sending rate
	void	start();                // start the sender
	void	stop();                 // stop the sender
	void	verify_switch(hdr_JetMax*, double); // manage bottleneck switching
protected:
	JetMaxSendTimer send_timer_;    // timer of scheduling out-going packet
	int		max_loss_seqno_;		// max loss sequence number seen so far
	int		cur_pkt_seqno_;			// current packet sequence number
	int		current_ACK_pkt_seqno_;	// packet sequence number of the current ACK
	int		rid_t;                  // ID of the selected bottleneck router
	int		fid_;					// flow ID
	int		active_;				// have we shut down? 
	int		switch_in_progress_;    // is a bottleneck switching in progress?
	int		test_switch_;           // delay response to a detected bottleneck switching by 1 RTT + 2 delta
	int		test_in_progress_;      // test whether a bottleneck switching is transient or persistent
	int		is_confirmed_;          // whether a proposed rate is confirmed by the router, if so wait for one more delta
	int		interval_cnt_;          // the number of intervals to be delayed
	int		is_rtt_expired_;        // has one RTT elapsed since the last rate change or bottleneck switching
	int		seqno_last_rtt;         // sequence number of the first packet during the last RTT interval
	int		cur_interval_;          // used to infer packet loss sequence number
	int		cur_loss_seqno_;        // current packet loss sequence number
	int		max_pkt_seqno_;         // maximum loss sequence number seen so far
	int     stationary;             // whether reaches stationarity under the current bottleneck assignment
	double	init_rate_;				// initial sending rate
	double	rate_;					// current actual sending rate
	double	rate_proposed_;         // proposed rate
	double	pkt_loss_;				// packet loss
	double	ref_rate_;				// reference rate
	double	psize_;					// packet size
	double	psize_proposed_;        // proposed packet size
	double	tau_;                   // control gain
	double	rate_avg_;              // ideal fair rate at the bottleneck router
};
